// Desktop Station Turnout Decoder 
// 20170702 Watchdog 4sec Add(Yaasan)
// 20170802 Interval 500ms changes to 100ms
// By yaasan


#include "NmraDcc_v141.h"
#include <avr/wdt.h>
#include <avr/eeprom.h>	 //required by notifyCVRead() function if enabled below


#define MAX_SERVO 2

#define PIN_PWM1A 5
#define PIN_PWM1B 6
#define PIN_PWM2A 3
#define PIN_PWM2B 11

#define PIN_RELAY2   7
#define PIN_RELAY  8 //R2基板以降対応

#define PIN_LED_DIV 9
#define PIN_LED_STR 10
#define PIN_LED_RUN 4

#define PIN_POS1   A7 //Str, R2基板ではA7
#define PIN_POS2   A1 //Div, R2基板ではA1
#define PIN_SPD    A2 //R2基板ではA2

#define PIN_DCIN   2

#define PI_ADRBIT_0 1
#define PI_ADRBIT_1 0
#define PI_ADRBIT_2 A5
#define PI_ADRBIT_3 A4
#define PI_ADRBIT_4 A3
#define PI_ADRBIT_5 A2
#define PI_ADRBIT_6 A1
#define PI_ADRBIT_7 A0


#define COUNT_OUTPUT 2 //10d=1sec

/* DCC SpeedStep 28専用テーブル */
const uint8_t DCC_28SPEEDMAP[29] = {
0b00000,0b00010,0b10010,0b00011,
0b10011,0b00100,0b10100,0b00101,
0b10101,0b00110,0b10110,0b00111,
0b10111,0b01000,0b11000,0b01001,
0b11001,0b01010,0b11010,0b01011,
0b11011,0b01100,0b11100,0b01101,
0b11101,0b01110,0b11110,0b01111,0b11111};

struct SRV_PARAM
{
	int mOn;
	int mSpeed;
	int mDir;
};
SRV_PARAM *gServoParam = new SRV_PARAM[MAX_SERVO];

NmraDcc  Dcc ;
DCC_MSG  Packet ;

struct CVPair
{
  uint16_t  CV;
  uint8_t   Value;
};
  
CVPair FactoryDefaultCVs [] =
{
	{CV_ACCESSORY_DECODER_ADDRESS_LSB, 1},
	{CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},
	{CV_29_CONFIG, 0b10000000},
	{47, 0},    //  Address  1: 軸1と軸2は同期運転(アドレス同一)、0:軸1と軸2は個別運転(アドレスは連番)
	{2, 32},   //  Duty Ratio 100%=255d
	{3, 32},   //  出力時間
	{50, 0},   //  Ch.1 Status
	{51, 0},   //  Ch.2 Status
	{52, 0},   //  1stアドレスオフセット
	{53, 0},   //  2ndアドレス
	{54, 0},   //  2ndアドレスオフセット
};

uint8_t FactoryDefaultCVIndex = 0;
unsigned long gPreviousWorkTime = 0;
uint16_t gLocoAddress = 1;
uint16_t gLocoAddress2 = 0;

uint8_t gCV47_AddrType = 0;
uint16_t gCV52_AddrOffset = 0;
uint16_t gCV53_Addr2 = 0;
uint16_t gCV54_AddrOffset2 = 0;
uint8_t gCV2_AccVal = 0;
uint8_t gCV3_DecVal = 0;


long gSpeedBuf[2] = {0, 0};


//Task Schedule
unsigned long gPreviousL5 = 0;
unsigned long gPreviousL6 = 0;
unsigned long gPreviousL7 = 0;

//LED control
uint8_t gLedRunCtrl = 0;


uint16_t getMyAddr_Acc(void);
void SetRelay(uint8_t inState);
void SetRelay2(uint8_t inState);
void SetLED(uint8_t inState);
uint16_t getMyAddrFromSW();


/* Functions  */



void notifyCVResetFactoryDefault()
{
  FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs)/sizeof(CVPair);
};

const int DccAckPin = PIN_RELAY ;

// This function is called by the NmraDcc library when a DCC ACK needs to be sent
void notifyCVAck(void)
{
  Serial.println("notifyCVAck") ;
 
  analogWrite( PIN_PWM1A, 200 );
  analogWrite( PIN_PWM1B, 0 );
  delay( 6 );  
  analogWrite( PIN_PWM1A, 0 );
}

void setup()
{
	
	
	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		#ifdef DEBUG
		//if eeprom has 0xFF then assume it needs to be programmed
		Serial.println("CV Defaulting due to blank eeprom");
		#endif
		notifyCVResetFactoryDefault();
	}
	else
	{
		#ifdef DEBUG
		Serial.println("CV Not Defaulting");
		#endif
	}
	
	
	// Setup which External Interrupt, the Pin it's associated with that we're using and enable the Pull-Up 
	Dcc.pin(0, 2, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( 140, 10, FLAGS_OUTPUT_ADDRESS_MODE | FLAGS_DCC_ACCESSORY_DECODER, 0 );

	/* アドレス読み出し */
	gCV52_AddrOffset = Dcc.getCV( 52);
	
	if( gCV52_AddrOffset > 7)
	{
		gCV52_AddrOffset = 0;
	}
	
	gLocoAddress = getMyAddrFromSW() + (gCV52_AddrOffset << 8);
	
	/* 2nd Address */
	gCV53_Addr2 = Dcc.getCV( 53);
	gCV54_AddrOffset2 = Dcc.getCV( 54);
	
	if( gCV54_AddrOffset2 > 7)
	{
		gCV54_AddrOffset2 = 0;
	}
	
	if( (gCV53_Addr2 == 0) && (gCV54_AddrOffset2 == 0))
	{
		/* CV53=CV54=0のときは連番アドレスを有効にする。どちらかに値が入っている場合はCh2アドレスが有効になる。 */
		gLocoAddress2 = gLocoAddress + 1;
	}
	else
	{
		gLocoAddress2 = gCV53_Addr2 + (gCV54_AddrOffset2 << 8);
		
		if( gLocoAddress2 > 2044)
		{
			/* アドレスを連番に強制変更 */
			gLocoAddress2 = gLocoAddress + 1;
		}
		
	}
	
	

	//アクセサリアドレス(下位2bitを考慮) 表示
	#ifdef DEBUG
	Serial.print("Acc Adr is ");
	Serial.println(gLocoAddress);  
	#endif
	
	gCV47_AddrType  = Dcc.getCV( 47);
	gServoParam[0].mOn = 0;
	gServoParam[0].mSpeed = 0;
	gServoParam[0].mDir = 0;
	gServoParam[1].mOn = 0;
	gServoParam[1].mSpeed = 0;
	gServoParam[1].mDir = 0;
	
	/* 完了 */
	
	Serial.begin(115200);
	
	#ifdef DEBUG
	Serial.println("Turnout decoder(Analog PWM)");
	#endif
	
	//初期位置へ
	
	//LED, Relay
	pinMode(PIN_RELAY, OUTPUT);
	pinMode(PIN_RELAY2, OUTPUT);
	pinMode(PIN_LED_DIV, OUTPUT);
	pinMode(PIN_LED_STR, OUTPUT);
	pinMode(PIN_LED_RUN, OUTPUT);
	
	SetRelay(gServoParam[0].mOn);
	SetRelay2(gServoParam[1].mOn);


	//PWM出力にセット
	pinMode(PIN_PWM1A, OUTPUT);
	pinMode(PIN_PWM1B, OUTPUT);
	pinMode(PIN_PWM2A, OUTPUT);
	pinMode(PIN_PWM2B, OUTPUT);

	#ifdef DEBUG
	Serial.println("Init Done");
	#endif
	
	gPreviousL7 = 0;
	gPreviousL6 = 0;
	gPreviousL5 = 0;
	
	/* 初期化 */
	gSpeedBuf[0] = 0;
	gSpeedBuf[1] = 0;
	
	
	wdt_enable(WDTO_4S); // WDT reset
}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();

	if( FactoryDefaultCVIndex && Dcc.isSetCVReady())
	{
		FactoryDefaultCVIndex--; // Decrement first as initially it is the size of the array 
		Dcc.setCV( FactoryDefaultCVs[FactoryDefaultCVIndex].CV, FactoryDefaultCVs[FactoryDefaultCVIndex].Value);
	}
	
	if( (millis() - gPreviousL6) >= 100)
	{
		//Reset task
		gPreviousL6 = millis();
		
		/* Reset watchdog timer. */
		wdt_reset(); 
		
		/* １軸目 */
		if( gServoParam[0].mOn == 1)
		{
			digitalWrite(PIN_LED_DIV, 0);
			
			gSpeedBuf[0] = ((236 * gSpeedBuf[0]) >> 8) + (20 * gServoParam[0].mSpeed);
			
			switch(gServoParam[0].mDir)
			{
			case 0:
				analogWrite(PIN_PWM1A, 255 - (gSpeedBuf[0] >> 8));
				analogWrite(PIN_PWM1B, 255);
				break;
				
			case 1:
				analogWrite(PIN_PWM1A, 255);
				analogWrite(PIN_PWM1B, 255 - (gSpeedBuf[0] >> 8));
				break;
				
			default:
				gServoParam[0].mOn = 0;
				break;
			}
			
		}
		else
		{
			digitalWrite(PIN_PWM1A, 1);
			digitalWrite(PIN_PWM1B, 1);
			digitalWrite(PIN_LED_DIV, 1);
		}
		
		/* ２軸目 */
		if( gServoParam[1].mOn == 1)
		{
			digitalWrite(PIN_LED_STR, 0);
			
			gSpeedBuf[1] = ((236 * gSpeedBuf[1]) >> 8) + (20 * gServoParam[1].mSpeed);
			
			switch(gServoParam[1].mDir)
			{
			case 0:
				analogWrite(PIN_PWM2A, 255 - (gSpeedBuf[1] >> 8));
				analogWrite(PIN_PWM2B, 255);
				
				break;
				
			case 1:
				analogWrite(PIN_PWM2A, 255);
				analogWrite(PIN_PWM2B, 255 - (gSpeedBuf[1] >> 8));
				
				break;
			default:
				gServoParam[1].mOn = 0;
				break;
			}
		}
		else
		{
			digitalWrite(PIN_PWM2A, 1);
			digitalWrite(PIN_PWM2B, 1);
			digitalWrite(PIN_LED_STR, 1);
		}
		
	}
	
	
	if( (millis() - gPreviousL7) >= 500)
	{
		//Reset task
		gPreviousL7 = millis();
		
		if( ( gServoParam[1].mOn == 1) || ( gServoParam[0].mOn == 1))
		{
			SetRunLED(0);
		}
		else
		{
			//LED control
			gLedRunCtrl++;
			
			if( gLedRunCtrl > 1)
			{
				SetRunLED(0);
				gLedRunCtrl = 0;
			}
			else
			{
				SetRunLED(1);
			}
		}
		
	}	
	
}


void exec_function (int function, int inDir, int inSpeed)
{
	
	if( function == 1)
	{
		
		if( inSpeed > 0)
		{
			gServoParam[0].mOn = 1;
		}
		else
		{
			gServoParam[0].mOn = 0;
		}
		
		gServoParam[0].mSpeed = inSpeed;
		gServoParam[0].mDir = inDir;
	}
	
	if( function == 2)
	{	
		if( inSpeed > 0)
		{
			gServoParam[1].mOn = 1;
		}
		else
		{
			gServoParam[1].mOn = 0;
		}
		
		gServoParam[1].mSpeed = inSpeed;
		gServoParam[1].mDir = inDir;
	}
	
}


void SetRelay(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_RELAY, 0);
		break;
	case 1:
		digitalWrite(PIN_RELAY, 1);
		break;
	default:
		digitalWrite(PIN_RELAY, 0);
		break;
	}	
}

void SetRelay2(uint8_t inState)
{
	
	switch( inState)
	{
	case 0:
		digitalWrite(PIN_RELAY2, 0);
		break;
	case 1:
		digitalWrite(PIN_RELAY2, 1);
		break;
	default:
		digitalWrite(PIN_RELAY2, 0);
		break;
	}	
}

void SetRunLED(uint8_t inState)
{
	digitalWrite(PIN_LED_RUN, inState);
}


uint16_t getMyAddrFromSW()
{
	uint16_t aAddr = 0;
	
	//Address DIP Switch
	pinMode(PI_ADRBIT_0,INPUT);
	pinMode(PI_ADRBIT_1,INPUT);	
	pinMode(PI_ADRBIT_2,INPUT);	
	pinMode(PI_ADRBIT_3,INPUT);	
	pinMode(PI_ADRBIT_4,INPUT);	
	pinMode(PI_ADRBIT_5,INPUT);	
	pinMode(PI_ADRBIT_6,INPUT);	
	pinMode(PI_ADRBIT_7,INPUT);	
	
	digitalWrite(PI_ADRBIT_0, HIGH);
	digitalWrite(PI_ADRBIT_1, HIGH);
	digitalWrite(PI_ADRBIT_2, HIGH);
	digitalWrite(PI_ADRBIT_3, HIGH);
	digitalWrite(PI_ADRBIT_4, HIGH);
	digitalWrite(PI_ADRBIT_5, HIGH);
	digitalWrite(PI_ADRBIT_6, HIGH);
	digitalWrite(PI_ADRBIT_7, HIGH);
	
	
	byte aBit0 = NotByte(digitalRead(PI_ADRBIT_0));
	byte aBit1 = NotByte(digitalRead(PI_ADRBIT_1));
	byte aBit2 = NotByte(digitalRead(PI_ADRBIT_2));
	byte aBit3 = NotByte(digitalRead(PI_ADRBIT_3));
	byte aBit4 = NotByte(digitalRead(PI_ADRBIT_4));
	byte aBit5 = NotByte(digitalRead(PI_ADRBIT_5));
	byte aBit6 = NotByte(digitalRead(PI_ADRBIT_6));
	byte aBit7 = NotByte(digitalRead(PI_ADRBIT_7));
	
	
	aAddr += aBit7;
	aAddr += aBit6 << 1;
	aAddr += aBit5 << 2;
	aAddr += aBit4 << 3;
	aAddr += aBit3 << 4;
	aAddr += aBit2 << 5;
	aAddr += aBit1 << 6;
	aAddr += aBit0 << 7;
	
	return aAddr;
	
}

byte NotByte(byte inByte)
{
	if( inByte == 0)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

/* 全DCCメッセージを受信 */
void notifyDccMsg( DCC_MSG * Msg)
{
	uint8_t aErrorbyte = 0;
	uint16_t aLocoAddr = 0;
	uint16_t aAccAddr = 0;


	/* エラーチェック */
	for(uint8_t i = 0; i < Msg->Size; i++)
	{
		aErrorbyte = aErrorbyte ^ Msg->Data[i];
	}

	/* エラーチェック */
	if( aErrorbyte != 0)
	{
		
		return;
	}
	else
	{
		
	}
	
	/* プリアンブル数計算 */
	if( Msg->PreambleBits < 10)
	{
		return;
	}
		
	/* パケットチェック */
	switch(Msg->Data[0])
	{
	case 0x00:
		//Broadcast
		break;
		
	case 0xFF:
		//Idle packet
		break;
		
	default:
		
		if( Msg->Data[0] <= 127)
		{
			/* SHORTアドレスロコの操作 */
			aLocoAddr = Msg->Data[0];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[1], Msg->Data[2]);
			
			
		}
		else if(( Msg->Data[0] >= 128) && ( Msg->Data[0] <= 191))
		{
			/* ポイント操作 */
		}
		else if(( Msg->Data[0] >= 192) && ( Msg->Data[0] <= 231))
		{
			/* LONGアドレスロコの操作 */
			aLocoAddr = ((Msg->Data[0] - 192) << 8) + Msg->Data[1];
			
			decodeLocoPacket(aLocoAddr, Msg->Data[2], Msg->Data[3]);
		}
		else if(( Msg->Data[0] >= 232) && ( Msg->Data[0] <= 254))
		{
			/* 予約 */
			
		}
		
		
		
		break;
		
	}
	
}

void decodeLocoPacket(word inLocoAddr, uint8_t inPacket1, uint8_t inPacket2)
{
	uint8_t aSpeed;
	uint8_t aDir;
	
	switch(inPacket1 >> 5)
	{
	case 0b000:
		break;
	
	case 0b001:
		
		switch(inPacket1 & 0b11111)
		{
			case 31:
				
				aSpeed = inPacket2 & 0b01111111;
				
			/* リスト更新 */
				updateDccSpeed( inLocoAddr, aSpeed, (inPacket2 >> 7) & 1, 127 );
			break;
		}
		break;

	case 0b010:
		//REV
			
		aSpeed = inPacket1 & 0b11111;
		
		for(int i = 0; i < 29; i++)
		{
			if( aSpeed == DCC_28SPEEDMAP[i])
			{
				/* リスト更新 */
				updateDccSpeed( inLocoAddr, i, 0, 28 );
				
				break;
			}
		}
		
		break;
	case 0b011:
		//FWD
			
		aSpeed = inPacket1 & 0b11111;
		
		for(int i = 0; i < 29; i++)
		{
			if( aSpeed == DCC_28SPEEDMAP[i])
			{
				/* リスト更新 */
				updateDccSpeed( inLocoAddr, i, 0, 28 );
				
				break;
			}
		}
		
		break;
		
	case 0b100:
		//bitWrite( gLocTable[aIndex].mFunction, 0, (inPacket1 >> 4) & 1);
		break;
		
	case 0b101:

		break;
		
	case 0b110:
		break;
		
	}
	

	
}


void updateDccSpeed( uint16_t inAddr, uint8_t inSpeed, uint8_t inDir, uint8_t inSpeedSteps )
{
	
	uint16_t aSpeed;
	uint16_t aLocAdress = inAddr;
	
	
	if ( inSpeed > 1)
	{
		aSpeed = (inSpeed - 1) * 255 / inSpeedSteps;
	}
	else
	{
		aSpeed = 0;
	}
	
	//アドレスチェック
	if( gLocoAddress == aLocAdress)
	{
		
		exec_function( 1,  inDir, aSpeed);
		
		/* ２軸目も同時に動かす */
		if( gCV47_AddrType == 1)
		{
			exec_function( 2,  inDir, aSpeed);
		}
		
	}
	else if( ((gLocoAddress2) == aLocAdress) && (gCV47_AddrType == 0) )
	{
		exec_function( 2,  inDir, aSpeed);
	}
	else
	{
		return;
	}
}

